/***********************************************************
 * $Id: ClassInfo.java 265 2012-12-15 12:42:36Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 02.02.2011
 * 
 ***********************************************************/

package org.clazzes.remoting.maven;

import japa.parser.ast.CompilationUnit;
import japa.parser.ast.ImportDeclaration;
import japa.parser.ast.body.BodyDeclaration;
import japa.parser.ast.body.EnumConstantDeclaration;
import japa.parser.ast.body.EnumDeclaration;
import japa.parser.ast.body.FieldDeclaration;
import japa.parser.ast.body.MethodDeclaration;
import japa.parser.ast.body.Parameter;
import japa.parser.ast.body.TypeDeclaration;
import japa.parser.ast.body.VariableDeclarator;
import japa.parser.ast.type.ClassOrInterfaceType;
import japa.parser.ast.type.PrimitiveType;
import japa.parser.ast.type.ReferenceType;
import japa.parser.ast.type.Type;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.clazzes.util.reflect.PropertyHelper;

/**
 * Information on a class, which is possibly retrieved from
 * the java source.
 */
public class ClassInfo {

	private final Class<?> clazz;
	private final FieldInfo[] declaredFields;
	private final MethodInfo[] declaredMethods;
	private final EnumConstantInfo[] enumConstants;
	private final Map<Method,MethodInfo> declaredMethodsByMethod;
	private final String   javaDoc;
	
	private static boolean typeEquals(Type t, Class<?> pt, CompilationUnit cu) {
		
		if (t instanceof PrimitiveType) {
			
			PrimitiveType prim = (PrimitiveType)t;
			
			switch(prim.getType()) {
			case Boolean: return (pt == Boolean.TYPE);
			case Byte:    return (pt == Byte.TYPE);
			case Char:    return (pt == Character.TYPE);
			case Double:  return (pt == Double.TYPE);
			case Float:   return (pt == Float.TYPE);
			case Int:     return (pt == Integer.TYPE);
			case Long:    return (pt == Long.TYPE);
			case Short:   return (pt == Short.TYPE);
			}
		}
		else if (t instanceof ClassOrInterfaceType) {
			
			ClassOrInterfaceType ct = (ClassOrInterfaceType) t;
			
			// Remark: This code does not work for types with ct.getScope() != null.
			//         This is not considered a major issue, because scope declarations
			//         in remote interfaces are rarely used.
			
			if (cu.getImports() != null) {
				for (ImportDeclaration imp : cu.getImports()) {

					if (imp.getName().getName().equals(ct.getName())) {
					
						return imp.getName().toString().equals(pt.getName());
					}
				}
			}
			
			String pkgNm = cu.getPackage().getName().toString() + "." + ct.getName();
			
			if (pkgNm.equals(pt.getName())) return true;
			
			pkgNm ="java.lang."+ct.getName();
			
			return pkgNm.equals(pt.getName());
		}
		else if (t instanceof ReferenceType) {
			
			ReferenceType rt = (ReferenceType)t;
												
			Class<?> ct = pt.getComponentType();
			
			if (ct == null) {
				
				if (rt.getArrayCount() == 0)
					return typeEquals(rt.getType(),pt,cu);
			}
			else {
				int depth = 0;
				
				while (pt.getName().charAt(depth) == '[') {
					++depth;
				}
				
				if (depth == rt.getArrayCount())
					return typeEquals(rt.getType(),ct,cu);
			}
		}
		
		return false;
	}
	
	private final void fillInFieldsAndMethodsSansSource() {
		
		Method[] mm = this.clazz.getDeclaredMethods();
		
		for (int i=0;i<mm.length;++i) {
			this.declaredMethods[i] = new MethodInfo(mm[i]);
		}
			
		Field[] ff = this.clazz.getDeclaredFields();
			
		for (int i=0;i<ff.length;++i) {
			this.declaredFields[i] = new FieldInfo(ff[i]);
		}
		
		Object [] ec = this.clazz.getEnumConstants();
		
		if (ec != null) {
			for (int i=0;i<ec.length;++i) {
				this.enumConstants[i] = new EnumConstantInfo(ec[i]);
			}
		}
	}
	
	private static final Map<Method,MethodInfo> buildMethodInfoMap(MethodInfo[] infos) {
		
		Map<Method,MethodInfo> ret = new HashMap<Method,MethodInfo>(infos.length);
		
		for (MethodInfo info : infos) {
		
			ret.put(info.getMethod(),info);
		}
		return ret;
	}
	
	public ClassInfo(Class<?> clazz) {
		
		this.clazz = clazz;
		this.declaredFields = new FieldInfo[clazz.getDeclaredFields().length];
		this.declaredMethods = new MethodInfo[clazz.getDeclaredMethods().length];
		Object [] ec = this.clazz.getEnumConstants();		
		if (ec == null) {
			this.enumConstants = null;
		}
		else {
			this.enumConstants = new EnumConstantInfo[ec.length];
		}
		this.fillInFieldsAndMethodsSansSource();
		this.declaredMethodsByMethod = buildMethodInfoMap(this.declaredMethods);
		this.javaDoc = null;
	}
	
	public ClassInfo(Class<?> clazz, CompilationUnit cu) {

		this.clazz = clazz;
		this.declaredFields = new FieldInfo[clazz.getDeclaredFields().length];
		this.declaredMethods = new MethodInfo[clazz.getDeclaredMethods().length];
		Object [] ec = this.clazz.getEnumConstants();		
		if (ec == null) {
			this.enumConstants = null;
		}
		else {
			this.enumConstants = new EnumConstantInfo[ec.length];
		}

		List<TypeDeclaration> tds = cu.getTypes();
		TypeDeclaration td = null;
		
		for (TypeDeclaration _td: tds) {
			
			if (_td.getName().equals(clazz.getSimpleName()) &&
			        cu.getPackage().getName().toString().equals(clazz.getPackage().getName())) {
				td = _td;
				break;
			}
		}
		
		if (td != null) {
			
			if (td.getJavaDoc() != null) {
				this.javaDoc = td.getJavaDoc().getContent();
			}
			else {
				this.javaDoc = null;
			}

			List<BodyDeclaration> bds = td.getMembers();
			Method[] mm = this.clazz.getDeclaredMethods();
			
			for (int i=0;i<mm.length;++i) {
				
				String javaDoc = null;
				String[] parameterNames = null;
				
				if (bds != null) {
					for (BodyDeclaration bd : bds) {
					
						if (bd instanceof MethodDeclaration) {
							
							MethodDeclaration _md = (MethodDeclaration)bd;
							
							if (_md.getName().equals(mm[i].getName())) {
								
								List<Parameter> pp = _md.getParameters();
								int pc = pp == null ? 0 : pp.size();
								
								if (pc == mm[i].getParameterTypes().length) {
									
									String[] _parameterNames = new String[pc];
									
									for (int j=0;j<pc;++j) {
										
										Parameter p = _md.getParameters().get(j);
										Class<?> pt = mm[i].getParameterTypes()[j];
										
										if (typeEquals(p.getType(),pt,cu)) {
											_parameterNames[j] = p.getId().getName();
										}
										else {
											_parameterNames = null;
											break;
										}
										
									}
									if (_parameterNames != null) {
										parameterNames = _parameterNames;
										if (_md.getJavaDoc() != null)
											javaDoc = _md.getJavaDoc().getContent();
										break;
									}
								}
							}
						}
					}
				}
				
				this.declaredMethods[i] = new MethodInfo(mm[i],parameterNames,javaDoc);
			}
				
			Field[] ff = this.clazz.getDeclaredFields();
				
			for (int i=0;i<ff.length;++i) {
				
				FieldDeclaration fd = null;
				
				if (bds != null) {
					for (BodyDeclaration bd : bds) {
	
						if (bd instanceof FieldDeclaration) {

							FieldDeclaration _fd = (FieldDeclaration)bd;
							
							if (_fd.getModifiers() == ff[i].getModifiers()) {
								
								List<VariableDeclarator> vars = _fd.getVariables();
								
								if (vars != null) {
									
									for (VariableDeclarator v:vars) {
										if (v.getId().getName().equals(ff[i].getName())) {
											fd = _fd;
											break;
										}
									}
								}
							}
							if (fd != null) break;
						}
					}
				}
				
				if (fd == null || fd.getJavaDoc() == null)
					this.declaredFields[i] = new FieldInfo(ff[i]);
				else
					this.declaredFields[i] = new FieldInfo(ff[i],fd.getJavaDoc().getContent());
			}

			if (ec != null) {
				
				List<EnumConstantDeclaration> ee = null;
				
				if (td instanceof EnumDeclaration) {
					
					ee = ((EnumDeclaration)td).getEntries();
				}
				
				for (int i=0;i<ec.length;++i) {
					
					EnumConstantDeclaration ecd = null;
					
					if (ee != null) {
						for (EnumConstantDeclaration _ecd : ee) {
							
							if (_ecd.getName().equals(ec[i].toString())) {
								
								ecd = _ecd;
								break;
							}
						}
					}
					
					if (ecd == null || ecd.getJavaDoc() == null)
						this.enumConstants[i] = new EnumConstantInfo(ec[i]);
					else
						this.enumConstants[i] = new EnumConstantInfo(ec[i],ecd.getJavaDoc().getContent());
				}
			}
		}
		else {
			this.fillInFieldsAndMethodsSansSource();
			this.javaDoc = null;
		}
		this.declaredMethodsByMethod = buildMethodInfoMap(this.declaredMethods);
	}
	
	public String getName() {
		return this.clazz.getName();
	}
	
	public int getModifiers() {
		return this.clazz.getModifiers();
	}
	
	public boolean isEnum() {
		return this.clazz.isEnum();
	}
	
	public Class<?> getSuperclass() {
		return this.clazz.getSuperclass();
	}
	
	public Class<?> getClazz() {
		return this.clazz;
	}

	public FieldInfo[] getDeclaredFields() {
		return this.declaredFields;
	}

	public MethodInfo[] getDeclaredMethods() {
		return this.declaredMethods;
	}

	public EnumConstantInfo[] getEnumConstants() {
		return this.enumConstants;
	}

	public MethodInfo getDeclaredMethod(String name, Class<?>... parameterTypes) throws SecurityException, NoSuchMethodException {
		
		return this.declaredMethodsByMethod.get(this.clazz.getDeclaredMethod(name, parameterTypes));
	}

	public MethodInfo getGetterMethod(FieldInfo field) throws SecurityException, NoSuchMethodException {
		
		if (field.getType().equals(boolean.class))
			return this.getDeclaredMethod(PropertyHelper.getBooleanGetterName(field.getName()));
		else
			return this.getDeclaredMethod(PropertyHelper.getGetterName(field.getName()));
	}

	public MethodInfo getSetterMethod(FieldInfo field) throws SecurityException, NoSuchMethodException {
		
		return this.getDeclaredMethod(PropertyHelper.getSetterName(field.getName()),field.getType());
	}

	public String getJavaDoc() {
		return this.javaDoc;
	}

	public String getSimpleName() {
		return this.clazz.getSimpleName();
	}

	public Class<?>[] getInterfaces() {
		
		return this.clazz.getInterfaces();
	}
}
