/***********************************************************
 * $Id: GenCxxMojo.java 300 2014-01-08 09:02:22Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 24.07.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.maven;

import japa.parser.JavaParser;
import japa.parser.ParseException;
import japa.parser.ast.CompilationUnit;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.rmi.Remote;
import java.rmi.server.UID;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.UUID;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.clazzes.remoting.annotations.OldSerialVersionUID;
import org.clazzes.remoting.annotations.OldSerialVersionUIDs;
import org.clazzes.util.datetime.ISO8601Format;
import org.clazzes.util.datetime.UtcTimestamp;
import org.clazzes.util.reflect.PackageHelper;
import org.clazzes.util.reflect.PropertyHelper;

/**
 * This mojo generates POCOs from the classes of one or more java packages.
 * 
 * @goal gencxx
 * @phase generate-sources
 */
public class GenCxxMojo extends AbstractMojo {

    /**
     * target directory for generated C++ source files.
     * 
     * @parameter default-value="${basedir}/src/main/c"
     */
    private File targetDirectory ;

    /**
	 * Names of classes which should be processed one by one.
	 * 
	 * @parameter
	 */
	private String[] classes; 
	
	/**
     * Names of classes whose packages should be processed as a whole.
     *
     * @parameter
     */
    private String[] packageClasses;

    /**
	 * Names of classes which should be treated as remote interfaces.
	 * if not specified, interfaces, which extend {@link Remote} are
	 * treated as remote interfaces.
	 * 
	 * @parameter
	 */
	private String[] remoteInterfaces; 

	/**
     * The target namespace of the generated C++ classes. If not set,
     * the namespace will be generated from the java package name.
     * Nested namespaces must be given as in dot-separated notation.  
     * 
     * @parameter
     */
    private String targetNamespace;
    
    /**
     * An additional ABI tag to include in the class declaration. This tag
     * may include compiler attributes like Windows DLL import/export attributes.
     * You may specify the parameter {@link #abiInclude} alongside with
     * this parameter in order to include an additional header file defining the
     * ABI tag, if this tag is a preprocessor define.
     * 
     * @parameter
     */
    private String abiTag;
    
    /**
     * An additional header to include inside the generated header file, which usually
     * declares the {@link #abiTag} used.
     * 
     * @parameter
     */
    private String abiInclude;
   
    /**
     * The generator aka the plugin name and the plugin version. This
     * is included in the heading comment of generated C++ files.
     */
    private String generator; 
    
    /**
     * Encoding to use, when parsing source files for JavaDoc and parameter names.
     * 
     * @parameter default-value="UTF-8"
     */
    private String sourceEncoding;

    /**
     * The fancybase ABI to generate code for. 
     * If set to <code>3</code> the field name based ABI of
     * fancybase 3 is used.
     * 
     * If set to <code>0</code>, omni-ABI code is generated, which
     * uses the FANYCLIBS_ABI define to distinguish between the
     * fancybase2 and fancybase3 ABI. 
     * 
     * @parameter default-value=0
     */
    private int fancybaseAbi;
    
    private static final Map<Class<?>,String> primitiveTypes;
    private static final Map<Class<?>,String> primitiveReadFns;
    private static final Map<Class<?>,String> primitiveArrayReadFns;
    private static final Map<Class<?>,String> primitiveWriteFns;
    private static final Map<Class<?>,String> primitiveToObjs;
    private static final Map<Class<?>,String> primitiveFromObjs;
    private static final Map<Class<?>,String> knownIncludes;
  
    static {
    	
    	primitiveTypes = new HashMap<Class<?>, String>();
    	
    	primitiveTypes.put(boolean.class,"bool");
    	primitiveTypes.put(char.class,"wchar_t");
    	primitiveTypes.put(byte.class,"int8_t");
    	primitiveTypes.put(short.class,"int16_t");
    	primitiveTypes.put(int.class,"int32_t");
    	primitiveTypes.put(long.class,"int64_t");
    	primitiveTypes.put(float.class,"float");
    	primitiveTypes.put(double.class,"double");
    	primitiveTypes.put(void.class,"void");
    	
    	primitiveReadFns = new HashMap<Class<?>, String>();
    	
    	primitiveReadFns.put(boolean.class,"readBoolean");
    	primitiveReadFns.put(char.class,"readChar");
    	primitiveReadFns.put(byte.class,"readByte");
    	primitiveReadFns.put(short.class,"readShort");
    	primitiveReadFns.put(int.class,"readInt");
    	primitiveReadFns.put(long.class,"readLong");
    	primitiveReadFns.put(float.class,"readFloat");
    	primitiveReadFns.put(double.class,"readDouble");
    	
    	primitiveArrayReadFns = new HashMap<Class<?>, String>();
    	
    	primitiveArrayReadFns.put(boolean.class,"readBooleanArray");
    	primitiveArrayReadFns.put(char.class,"readCharArray");
    	primitiveArrayReadFns.put(byte.class,"readByteArray");
    	primitiveArrayReadFns.put(short.class,"readShortArray");
    	primitiveArrayReadFns.put(int.class,"readIntArray");
    	primitiveArrayReadFns.put(long.class,"readLongArray");
    	primitiveArrayReadFns.put(float.class,"readFloatArray");
    	primitiveArrayReadFns.put(double.class,"readDoubleArray");
    	
    	primitiveWriteFns = new HashMap<Class<?>, String>();
    	
    	primitiveWriteFns.put(boolean.class,"writeBoolean");
    	primitiveWriteFns.put(char.class,"writeChar");
    	primitiveWriteFns.put(byte.class,"writeByte");
    	primitiveWriteFns.put(short.class,"writeShort");
    	primitiveWriteFns.put(int.class,"writeInt");
    	primitiveWriteFns.put(long.class,"writeLong");
    	primitiveWriteFns.put(float.class,"writeFloat");
    	primitiveWriteFns.put(double.class,"writeDouble");
    	
    	primitiveToObjs = new HashMap<Class<?>, String>();
    	
    	primitiveToObjs.put(boolean.class,"Fancy::Boolean::newInstance");
    	primitiveToObjs.put(char.class,"Fancy::Character::newInstance");
    	primitiveToObjs.put(byte.class,"Fancy::Byte::newInstance");
    	primitiveToObjs.put(short.class,"Fancy::Short::newInstance");
    	primitiveToObjs.put(int.class,"Fancy::Integer::newInstance");
    	primitiveToObjs.put(long.class,"Fancy::Long::newInstance");
    	primitiveToObjs.put(float.class,"Fancy::Float::newInstance");
    	primitiveToObjs.put(double.class,"Fancy::Double::newInstance");
    	
    	primitiveFromObjs = new HashMap<Class<?>, String>();
    	
    	primitiveFromObjs.put(boolean.class,"Fancy::Boolean::toBoolean");
    	primitiveFromObjs.put(char.class,"Fancy::Character::toCharacter");
    	primitiveFromObjs.put(byte.class,"Fancy::Byte::toByte");
    	primitiveFromObjs.put(short.class,"Fancy::Short::toShort");
    	primitiveFromObjs.put(int.class,"Fancy::Integer::toInteger");
    	primitiveFromObjs.put(long.class,"Fancy::Long::toLong");
    	primitiveFromObjs.put(float.class,"Fancy::Float::toFloat");
    	primitiveFromObjs.put(double.class,"Fancy::Double::toDouble");
    	
    	knownIncludes = new HashMap<Class<?>, String>();
    	knownIncludes.put(Boolean.class,"fancynumbers.h");
    	knownIncludes.put(Byte.class,"fancynumbers.h");
    	knownIncludes.put(Character.class,"fancynumbers.h");
    	knownIncludes.put(Short.class,"fancynumbers.h");
    	knownIncludes.put(Integer.class,"fancynumbers.h");
    	knownIncludes.put(Long.class,"fancynumbers.h");
    	knownIncludes.put(Float.class,"fancynumbers.h");
    	knownIncludes.put(Double.class,"fancynumbers.h");
    	knownIncludes.put(String.class,"fancystring.h");
    	knownIncludes.put(UID.class,"fancyuid.h");
    	knownIncludes.put(UUID.class,"fancyuuid.h");
     	knownIncludes.put(UtcTimestamp.class,"fancyutctimestamp.h");
     	knownIncludes.put(List.class,"fancylist.h");
    	knownIncludes.put(Set.class,"fancyset.h");
    	knownIncludes.put(Map.class,"fancymap.h");	
    	knownIncludes.put(Throwable.class,"fancythrowable.h");	
    	knownIncludes.put(Exception.class,"fancythrowable.h");	
       	knownIncludes.put(RuntimeException.class,"fancythrowable.h");
       	knownIncludes.put(Object.class,"");
    }

    private ClassInfo getClassInfo(Class<?> clazz) {
    	
    	String resource = clazz.getName().replace('.','/') + ".java";
    	
    	InputStream source =
    		GenCxxMojo.class.getClassLoader().getResourceAsStream(resource);
    	
    	if (source != null) {

    		try {
    			
				CompilationUnit cu = JavaParser.parse(source,this.sourceEncoding);
				
				this.getLog().debug("Resource ["+resource+"] found, JavaDocs and parameter names are available.");
				
				return new ClassInfo(clazz,cu);
				
			} catch (ParseException e) {
				this.getLog().warn("Error parsing resource ["+resource+"], JavaDocs and parameter names will not be available",e);
			}
    	}
    	
    	this.getLog().debug("Resource ["+resource+"] not found, JavaDocs and parameter names will not be available.");
    	
    	return new ClassInfo(clazz);
    }
       
    private static void quoteCString(PrintWriter pw, String s) throws UnsupportedEncodingException {
    	
    	pw.append("\"");
    	
    	byte[] bytes = s.getBytes("UTF-8");
    	
    	for (byte b : bytes) {
    	
    		if (b < 0x20 || b > 126) {
    			
    			pw.printf(Locale.ENGLISH,"\\%03o",((int)b)&0xff);
    		}
    		else {
    			pw.append((char)b);
    		}
    	}
    	
    	pw.append("\"");
    }
    
    private static void printIndented(PrintWriter pw, int indent, String line) {
    
    	for (int i=0;i<indent;++i)
    		pw.print(' ');
    	pw.println(line);
    }
    
    
    private static void printJavaDoc(PrintWriter pw, int indent, String javaDoc) {
    	
    	if (javaDoc == null) return;
    	
    	printIndented(pw, indent,"/**");
    	
    	int npos,pos = 0;
    	
    	while((npos = javaDoc.indexOf('\n',pos)) >= pos) {
    		
    		int next = npos+1;
    		
    		if (npos > pos && javaDoc.charAt(npos-1) == '\r')
    			--npos;
    			
    		while (pos < npos && Character.isWhitespace(javaDoc.charAt(pos)))
    			++pos;
    		
    		if (npos > 0)
    			printIndented(pw,indent+1,javaDoc.substring(pos,npos));
    		
    		pos = next;
    	}
    	
		while (pos < javaDoc.length() && Character.isWhitespace(javaDoc.charAt(pos)))
			++pos;
		
		if (pos<javaDoc.length())
			printIndented(pw,indent+1,javaDoc.substring(pos));
		
		printIndented(pw,indent+1,"*/");
    }
    
    private static void writeHiddenClassHeader(PrintWriter pw) {
    	
    	pw.println("class");
    	pw.println("#if ( defined __GNUG__ ) && ! ( defined WIN32 )");
    	pw.println("__attribute__((visibility(\"hidden\")))");
    	pw.println("#endif");
    }
    
    private static long fetchSerialVerionUID(Class<?> clazz) {
        
    	try {
    		
            Field sv = clazz.getDeclaredField("serialVersionUID");
            
            sv.setAccessible(true);
            
            if (Modifier.isStatic(sv.getModifiers()) &&
            		Modifier.isFinal(sv.getModifiers()))
            	return ((Long)sv.get(null)).longValue();
            
    	} catch (NoSuchFieldException e) {
           
    	} catch (Throwable e) {
    		throw new RuntimeException("Cannot fetch serialVersionUID of ["+clazz+"]",e);
    	}
        
    	return -1L;
    }
    
    private static boolean isVisitorAccept(MethodInfo m) {
    	
		return
		"accept".equals(m.getName())&&
		!Modifier.isStatic(m.getModifiers()) &&
		m.getParameterTypes().length == 1 &&
   		m.getParameterTypes()[0].isInterface() &&
   		m.getReturnType() == void.class;
    }
    
    private boolean isFancyClass(Class<?> c) {
    	
    	return
    	c==Boolean.class || c==Byte.class || c==Character.class ||
    	c==Float.class || c==Double.class || c==String.class ||
    	c==Short.class || c==Integer.class || c == Object.class ||
    	c==Long.class || c==UUID.class ||
    	c==URL.class || c==UtcTimestamp.class ||
    	c==UID.class || c==List.class ||
    	c==Set.class || c == Map.class||
    	c==Throwable.class || c == Exception.class || c == RuntimeException.class;
    }
    
    private String getInclude(Class<?> c) {
    	
    	if (c.isPrimitive()) return null;
    	String inc = knownIncludes.get(c);
    	
    	if (inc != null)
    		return inc.length() <= 0 ? null : inc;
    	
    	if (this.targetNamespace == null) {
        	
    		return c.getName().replace('.','_') + ".h";
    	}
    	else {
    
    		return this.targetNamespace.replace('.','_') + "_" + c.getSimpleName() + ".h";
    	}
    }
    
    private void addInclude(Set<String> includes, Class<?> c) {
    	
    	if (c.isArray()) {
    		includes.add("fancylist.h");
    		c=c.getComponentType();
    	}
    	
    	String inc = this.getInclude(c);
    	if (inc == null) return;
    	includes.add(inc);
    }
    
    private File getCxxFile(Class<?> c) {
    	
    	String cxxFileName;
    	
    	if (this.targetNamespace == null) {
    	
    		cxxFileName = c.getName().replace('.','_') + ".cpp";
    	}
    	else {
    
    		cxxFileName = this.targetNamespace.replace('.','_') + "_" + c.getSimpleName() + ".cpp";
    	}
    	
    	return new File(this.targetDirectory,cxxFileName);
    }
    	
    private File getHeaderFile(Class<?> c) {
    	
    	String headerFileName;
    	
    	if (this.targetNamespace == null) {
    	
    		headerFileName = c.getName().replace('.','_') + ".h";
    	}
    	else {
    
    		headerFileName = this.targetNamespace.replace('.','_') + "_" + c.getSimpleName() + ".h";
    	}
    	
    	return new File(this.targetDirectory,headerFileName);
    }
    	
    private String getNamespace(Class<?> c) {
    	
    	if (c.isPrimitive()) return null;
    	
    	String namespace;
       	
    	if (this.isFancyClass(c))
    		namespace = "Fancy";
    	else
    		namespace = this.targetNamespace;
    	
    	if (namespace == null) {
    		
    		int pos = c.getName().lastIndexOf('.');
    		if (pos >= 0)
       			namespace = c.getName().substring(0,pos);
    	}
    	
    	return namespace;
    }
    
    private int writeNamespace(PrintWriter pw, String namespace) {
    	
      	if (namespace == null || namespace.length() <= 0) return 0;

       	int dp = 0;
       	int namespaceDepth = 0;
       	
       	do {
       		
       		int ndp = namespace.indexOf('.',dp);
       		if (ndp < 0) ndp = namespace.length();
       		
       		if (namespaceDepth > 0) pw.print(' ');
       		 
       		pw.print("namespace " + namespace.substring(dp,ndp) + " {");
       		
       		++namespaceDepth;
       		dp = ndp + 1;
       		
       	} while (dp < namespace.length());
       	
       	pw.println();
       	
    	return namespaceDepth;
    }
    
    private int writeNamespace(PrintWriter pw, Class<?> c) {
    	
       	String namespace = this.getNamespace(c);
       	
       	return this.writeNamespace(pw,namespace);
    }
    
    private static void addFwdDecl(SortedMap<String,SortedSet<String>> fwdDeclsByNamespace, String namespace, String decl) {
    	
      	SortedSet<String> fwdDecls = fwdDeclsByNamespace.get(namespace);
    	if (fwdDecls == null) {
    		fwdDecls = new TreeSet<String>();
    		fwdDeclsByNamespace.put(namespace,fwdDecls);
    	}
    	
    	fwdDecls.add(decl);
    }
    
    private void addFwdDecl(SortedMap<String,SortedSet<String>> fwdDeclsByNamespace, Class<?> c) {
    	
    	if (c.isArray()) {
   			
   			if (c.getComponentType().isPrimitive()) {
   				addFwdDecl(fwdDeclsByNamespace,"Fancy","template<class T> class ArrayObject");
   			} else {
   				addFwdDecl(fwdDeclsByNamespace,"Fancy","template<class T> class ObjectArray");
   	   			this.addFwdDecl(fwdDeclsByNamespace,c.getComponentType());
   			}
   		}
   		else {  			
   			String namespace = this.getNamespace(c);
   			if (namespace == null) return;
			
   			String sn;
   			
			if (c == Exception.class || c == RuntimeException.class)
				sn =  "class Java" + c.getSimpleName();
			else
				sn = "class "+c.getSimpleName();

   			addFwdDecl(fwdDeclsByNamespace,namespace,sn);
  		}       		

    }
    
    private String getArgPfx(Class<?> c) {
    	
   		if (c.isArray()) {
   			
   			if (c.getComponentType().isPrimitive()) {
   				
       			return "const Fancy::ObjPtr<Fancy::ArrayObject<"+this.getCxxClassName(c.getComponentType())+"> > &";
       			
   			} else {
   				
       			return "const Fancy::ObjPtr<Fancy::ObjectArray<"+this.getCxxClassName(c.getComponentType())+"> > &";
   			}
   			
   		}
   		else if (c.isPrimitive()) {
   			
   			return this.getCxxClassName(c)+" ";
   		}
   		else {
   			return "const Fancy::ObjPtr<"+this.getCxxClassName(c)+"> &";
  		}
    }
    
    private String getRetVal(Class<?> c) {
    	
   		if (c.isArray()) {
   			
   			if (c.getComponentType().isPrimitive()) {
   				
       			return "Fancy::ObjPtr<Fancy::ArrayObject<"+this.getCxxClassName(c.getComponentType())+"> > ";
       			
   			} else {
   				
       			return "Fancy::ObjPtr<Fancy::ObjectArray<"+this.getCxxClassName(c.getComponentType())+"> > ";
   			}
   			
   		}
   		else if (c.isPrimitive()) {
   			
   			return this.getCxxClassName(c)+" ";
   		}
   		else {
   			return "Fancy::ObjPtr<"+this.getCxxClassName(c)+"> ";
  		}
    }
    
    private String getCxxClassName(Class<?> c) {
    	
    	String pn = primitiveTypes.get(c);
    	if (pn != null) return pn;
    	
    	if (this.isFancyClass(c)) {
    		
    		if (c == Exception.class || c == RuntimeException.class)
    			return "Fancy::Java" + c.getSimpleName();
    		else
    			return "Fancy::" + c.getSimpleName();
    	}
    	
    	if (this.targetNamespace == null) {
        	
    		return c.getName().replace(".","::");
    	}
    	else {
    		
    		return this.targetNamespace.replace(".","::") + "::" + c.getSimpleName();
    	}
    	
    }
    
    private String formatPrimitiveConstant(Field f) throws IllegalArgumentException, IllegalAccessException {
    
    	if (f.getType() == boolean.class) {
    		
    		if (f.getBoolean(null))
    			return "true";
    		else
    			return "false";
    	}
    	
    	if (f.getType() == byte.class) {  		
    		return "INT8_C("+f.getByte(null)+")";
    	}
    	
    	if (f.getType() == short.class) {  		
    		return "INT16_C("+f.getShort(null)+")";
    	}
    	
    	if (f.getType() == int.class) {  		
    		return Integer.toString(f.getInt(null));
    	}
    	
    	if (f.getType() == long.class) {  		
    		return "INT64_C("+f.getLong(null)+")";
    	}
   	
       	if (f.getType() == float.class) {  		
    		return f.getFloat(null)+"f";
    	}
   	
       	if (f.getType() == double.class) {  		
    		return Double.toString(f.getDouble(null));
    	}
   	
       	if (f.getType() == char.class) {  		
    		return "L'"+f.getChar(null)+"'";
    	}
   	
       	throw new IllegalArgumentException("Unsupported constant type ["+f.getType()+"].");
  }
    
    private void propagatePrimitiveFinalFields(PrintWriter pw, ClassInfo classInfo) throws IllegalArgumentException, IllegalAccessException {
    	
    	int count = 0;
    	
       	for (FieldInfo f : classInfo.getDeclaredFields()) {
       		
       		if (Modifier.isStatic(f.getModifiers()) &&
       				Modifier.isFinal(f.getModifiers()) &&
       				Modifier.isPublic(f.getModifiers()) &&
       				f.getType().isPrimitive()) {
       			
       			printJavaDoc(pw,2,f.getJavaDoc());
       			
       			pw.println("  static const "+this.getCxxClassName(f.getType())+" "+f.getName()+"="+this.formatPrimitiveConstant(f.getField())+";");
       			++count;
       		}
       	}
       	if (count > 0) pw.println();
    }
    	
    /**
     * @param c
     * @throws FileNotFoundException
     * @throws UnsupportedEncodingException
     * @throws IllegalArgumentException
     * @throws IllegalAccessException
     */
    private void processPOJOClass(ClassInfo c) throws FileNotFoundException, UnsupportedEncodingException, IllegalArgumentException, IllegalAccessException {
    
    	ISO8601Format fmt = new ISO8601Format(ISO8601Format.DATETIME_FORMAT);
    	boolean hasSuper = c.getSuperclass() != Object.class;
    	
    	File cxxFile = this.getCxxFile(c.getClazz());
    	File headerFile = this.getHeaderFile(c.getClazz());
    	String defn = headerFile.getName().toUpperCase(Locale.ENGLISH).replace('.','_');
    	
    	this.getLog().info("Writing header of POJO class ["+c.getName()+"] to ["+headerFile+"].");
    	
    	PrintWriter pw = new PrintWriter(headerFile,"utf-8");
       	
    	pw.println("//"); 
    	pw.println("// File: " + headerFile.getName()); 
    	pw.println("// generated: " + fmt.format(Calendar.getInstance())); 
    	pw.println("// generator: "+this.generator+"."); 
       	pw.println("// Note: Use 'mvn process-sources' to regenerate this file."); 
       	pw.println("//"); 
       	pw.println();
       	pw.println("#ifndef "+defn);
       	pw.println("#define "+defn+" 1");
       	pw.println();
       	
       	if (this.abiInclude != null)
       		pw.println("#include <"+this.abiInclude+">"); 
       	
       	if (hasSuper) {
       		pw.println("#include <"+this.getInclude(c.getSuperclass())+">");
       	}
       	else {
       		pw.println("#include <fancyserializable.h>"); 
       	}
       	
       	SortedMap<String,SortedSet<String>> fwdDeclsByNamespace = new TreeMap<String, SortedSet<String>>();
       	
       	int fieldCount = 0;
       	
       	// include fwd decls for field types.
       	for (FieldInfo f : c.getDeclaredFields()) {
       		
       		if (Modifier.isStatic(f.getModifiers()) ||
       				Modifier.isTransient(f.getModifiers()))
       			continue;
       		
       		++fieldCount;
       		this.addFwdDecl(fwdDeclsByNamespace,f.getType());
       	}

       	pw.println();
       	
       	// Add fwd decls for visitors.
       	for (MethodInfo m : c.getDeclaredMethods()) {
       		
       		if (!isVisitorAccept(m)) continue;
       		this.addFwdDecl(fwdDeclsByNamespace,m.getParameterTypes()[0]);
       	}

       	for(Map.Entry<String,SortedSet<String>> fwdDecl : fwdDeclsByNamespace.entrySet()) {
       		
       		int n = this.writeNamespace(pw,fwdDecl.getKey());
       		for (String decl : fwdDecl.getValue()) {
       			pw.println(decl+";");
       		}
       		for (int i=0;i<n;++i) {
        		pw.print("}");
        	}
       		pw.println();
       	}
       	
       	pw.println();
       	
       	String clsDecl;
       	
       	if (this.abiTag != null)
       		clsDecl = this.abiTag + " " + c.getSimpleName();
       	else
       		clsDecl = c.getSimpleName();
       	
       	int namespaceDepth = this.writeNamespace(pw,c.getClazz());
       	
       	printJavaDoc(pw,0,c.getJavaDoc());
       	
       	if (hasSuper) {
       		pw.println("class "+clsDecl+" : public " +
       				   this.getCxxClassName(c.getSuperclass()) + " {");
       	} else {
       		pw.println("class "+clsDecl+" : virtual public Fancy::Serializable {");
       	    
       	}
       	
       	pw.println("protected:");
       	pw.println("  "+c.getSimpleName()+"();");
       	pw.println("  virtual ~"+c.getSimpleName()+"();");
       	pw.println(); 
       	pw.println("public:");
       	pw.println("  typedef Fancy::ObjPtr<"+c.getSimpleName()+"> ptr_t;");
       	if (!Modifier.isAbstract(c.getModifiers()))
       		pw.println("  static ptr_t newInstance();");
       	pw.println();
       	
       	this.propagatePrimitiveFinalFields(pw,c);
       	
       	pw.println("  // implementation of Fancy::Serializable"); 
      	if (!Modifier.isAbstract(c.getModifiers())) {
      		pw.println("  static Fancy::Serializable::ptr_t newSerializableInstance();");
      	}
      	pw.println("  static int64_t getSerialVersionUID_s();");
      	pw.println("  static const char *getSerialID_s();");
      	pw.println("  int64_t getSerialVersionUID() const;");
      	pw.println("  const char *getSerialID() const;");
      	
      	if (hasSuper) {
      		pw.println("  static const char *getSuperSerialID_s();");
      		pw.println("  const char *getSuperSerialID() const;");
      	}
      	
      	if (fieldCount > 0 || !hasSuper) {
      		pw.println("  void writeFields(const Fancy::ObjPtr<Fancy::Serializer> &serializer) const;");
      		pw.println("  Fancy::Serializable::ptr_t readFields(const Fancy::ObjPtr<Fancy::Serializer> &serializer, int64_t serialVersionUID);");
      	}
       	pw.println(); 
       	
       	// go through fields and write getters/setters.
       	for (FieldInfo f : c.getDeclaredFields()) {
       		
       		if (Modifier.isStatic(f.getModifiers()) ||
       				Modifier.isTransient(f.getModifiers()))
       			continue;
       		
       		String getter = 
       			f.getType() == boolean.class ?
       			PropertyHelper.getBooleanGetterName(f.getName()):
                PropertyHelper.getGetterName(f.getName());
       					
       		String setter = PropertyHelper.getSetterName(f.getName());
 
       		String getterJavaDoc = null;
       		
       		try {
       			getterJavaDoc = c.getGetterMethod(f).getJavaDoc();
       		} catch (Throwable e) {
       		}
       		
       		String setterJavaDoc = null;
       		
       		try {
       			setterJavaDoc = c.getSetterMethod(f).getJavaDoc();
       		} catch (Throwable e) {
       		}
       		
   			printJavaDoc(pw,2,setterJavaDoc);
       		pw.println("  void " +setter +
   					"("+this.getArgPfx(f.getType())+f.getName()+");");
       		
       		
   			printJavaDoc(pw,2,getterJavaDoc);
  			pw.println("  "+this.getRetVal(f.getType()) + getter + "() const;");
       	}
       	
       	// Add visitor accept methods.
       	for (MethodInfo m : c.getDeclaredMethods()) {
       		
       		if (!isVisitorAccept(m)) continue;
       		
       		if (Modifier.isAbstract(m.getModifiers()))
       			pw.println("  virtual void accept("+this.getArgPfx(m.getParameterTypes()[0])+"v) = 0;");
       		else
       			pw.println("  void accept("+this.getArgPfx(m.getParameterTypes()[0])+"v);");
       	}
       	
        // Add private fields.
       	pw.println("private:");
       	for (FieldInfo f : c.getDeclaredFields()) {
       		
       		if (Modifier.isStatic(f.getModifiers()) ||
       				Modifier.isTransient(f.getModifiers()))
       			continue;
       		
   			printJavaDoc(pw,2,f.getJavaDoc());
       		if (f.getType().isArray()) {
       			
       			if (f.getType().getComponentType().isPrimitive()) {
       				
           			pw.println("  Fancy::ObjPtr<Fancy::ArrayObject<"+this.getCxxClassName(f.getType().getComponentType())+"> > "+
           					f.getName()+";");
       			} else {
       				
           			pw.println("  Fancy::ObjPtr<Fancy::ObjectArray<"+this.getCxxClassName(f.getType().getComponentType())+"> > "+
           					f.getName()+";");
       			}
       			
       		}
       		else if (f.getType().isPrimitive()) {
       			
       			pw.println("  "+this.getCxxClassName(f.getType())+" "+f.getName()+";");
       		}
       		else {
       			pw.println("  Fancy::ObjPtr<"+this.getCxxClassName(f.getType())+"> "+
       					f.getName()+";");
      		}
       	}

       	pw.println("};"); // end of class
    	
       	// end of namespace(s).
    	for (int i=0;i<namespaceDepth;++i) {
    		pw.print("}");
    	}
    	pw.println();
    	
    	// set the ObjPtr template to the same ABI visibility than the class itself.
    	if (this.abiTag != null) {
    		pw.println("namespace Fancy {");
    		pw.println("  template class "+this.abiTag+" ObjPtr<"+this.getCxxClassName(c.getClazz())+">;");
    		pw.println("}");
    	}

    	pw.println("#endif // "+defn);
    	pw.close();
    	
    	// write implementation file.
    	this.getLog().info("Writing implementation of POJO class ["+c.getName()+"] to ["+cxxFile+"].");
    	
    	pw = new PrintWriter(cxxFile,"utf-8");
       	
    	pw.println("//"); 
    	pw.println("// File: " + cxxFile.getName()); 
    	pw.println("// generated: " + fmt.format(Calendar.getInstance())); 
    	pw.println("// generator: "+this.generator+"."); 
       	pw.println("// Note: Use 'mvn process-sources' to regenerate this file."); 
       	pw.println("//"); 
       	pw.println();
       	
       	SortedSet<String> includes = new TreeSet<String>();
       	SortedMap<String,FieldInfo> fieldNames = new TreeMap<String,FieldInfo>();
       
       	includes.add("fancyserializer.h");
       		
       	this.addInclude(includes,c.getClazz()); 
       	
       	// include declarations for field types.
       	for (FieldInfo f : c.getDeclaredFields()) {
       		
       		if (Modifier.isStatic(f.getModifiers()) ||
       				Modifier.isTransient(f.getModifiers()))
       			continue;
       		
       		fieldNames.put(f.getName(),f);
       		
       		if (f.getType().isArray()) {
       			
       			if (f.getType().getComponentType().isPrimitive()) {
       				includes.add("fancyarray.h");
       			} else {
       				includes.add("fancylist.h");
       				this.addInclude(includes,f.getType().getComponentType());
       			}
       		}
       		else {
       			
       			this.addInclude(includes,f.getType());
       			
       			for (int i=0;i<f.getGenericParameterCount();++i) {
       				
       				this.addInclude(includes,f.getGenericParameter(i));
       			}
      		}       		
       	}

       	Map<String,List<Long>> fieldExclusionsByOSVUID = new HashMap<String, List<Long>>(fieldNames.size());
       	Map<Long,Long> superSVUIDs = null;
       	
       	{
       		OldSerialVersionUID osvuid = c.getClazz().getAnnotation(OldSerialVersionUID.class);
       		
       		if (osvuid != null) {
       			
       			String superId = osvuid.superSerialVersionUID();
       			
       			if (superId != null && !superId.isEmpty()) {
       				
       				if (superSVUIDs == null) {
       					superSVUIDs = new HashMap<Long, Long>();
       				}
       				
       				superSVUIDs.put(osvuid.id(),Long.valueOf(superId));
       			}
       			
       			for (String field : osvuid.skipFields()) {
       				
       				List<Long> ids = fieldExclusionsByOSVUID.get(field);
       				
       				if (ids == null) {
       					ids = new ArrayList<Long>();
       					fieldExclusionsByOSVUID.put(field,ids);
       				}
       				
       				ids.add(osvuid.id());
       			}
       		}
       	}
       	
       	OldSerialVersionUIDs osvuids = c.getClazz().getAnnotation(OldSerialVersionUIDs.class);
       	
       	if (osvuids != null) {
       		
       		for (OldSerialVersionUID  osvuid : osvuids.value()) {
       			
           		String superId = osvuid.superSerialVersionUID();
           		
           		if (superId != null && ! superId.isEmpty()) {
       				
           			if (superSVUIDs == null) {
       					superSVUIDs = new HashMap<Long, Long>();
       				}

       				superSVUIDs.put(osvuid.id(),Long.valueOf(superId));
           		}

           		for (String field : osvuid.skipFields()) {
           			
           			List<Long> ids = fieldExclusionsByOSVUID.get(field);
           			
           			if (ids == null) {
           				ids = new ArrayList<Long>();
               			fieldExclusionsByOSVUID.put(field,ids);
           			}

           			ids.add(osvuid.id());
           		}
       		}
       	}
       	
       	// include fwd decls for visitors.
       	for (MethodInfo m : c.getDeclaredMethods()) {
       		
       		if (!isVisitorAccept(m)) continue;
       		this.addInclude(includes,m.getParameterTypes()[0]);
       	}

       	for(String include : includes) {
       		
       		pw.println("#include <"+include+">");
       	}
       	
       	pw.println();
       	
       	namespaceDepth = this.writeNamespace(pw,c.getClazz());
       	
       	// ctor
       	pw.println(c.getSimpleName()+"::"+c.getSimpleName()+"()");
       	
       	int nfields = 0;
       	
       	// initialize primitive fields to zero.
       	for (FieldInfo f : c.getDeclaredFields()) {
       		
       		if (Modifier.isStatic(f.getModifiers()) ||
       				Modifier.isTransient(f.getModifiers()))
       			continue;
       		
       		if (!f.getType().isPrimitive()) continue;
       		
       		pw.print(nfields > 0 ? "," : " :");
       		pw.print(f.getName()+"(0)");
       		++nfields;
       	}
       	if (nfields > 0) pw.println(); 
       	pw.println("{}");
       	pw.println();
       	// dtor
       	pw.println(c.getSimpleName()+"::~"+c.getSimpleName()+"()");
      	pw.println("{}");
       	pw.println(); 
       	
      	if (!Modifier.isAbstract(c.getModifiers())) {
      		pw.println(c.getSimpleName()+"::ptr_t "+c.getSimpleName()+"::newInstance() {");
      		pw.println("  return new "+c.getSimpleName()+"();"); 
      		pw.println("}"); 
      		pw.println(); 
      		pw.println("Fancy::Serializable::ptr_t "+c.getSimpleName()+"::newSerializableInstance() {");
      		pw.println("  return new "+c.getSimpleName()+"();"); 
      		pw.println("}"); 
      		pw.println(); 
      	}
      	pw.println("int64_t "+c.getSimpleName()+"::getSerialVersionUID_s() {");
      	pw.println("  return INT64_C("+fetchSerialVerionUID(c.getClazz())+");"); 
      	pw.println("}"); 
      	pw.println(); 
      	pw.println("const char *"+c.getSimpleName()+"::getSerialID_s() {");
      	pw.println("  return \""+c.getName()+"\";"); 
      	pw.println("}"); 
      	pw.println(); 
      	pw.println("int64_t "+c.getSimpleName()+"::getSerialVersionUID() const {");
      	pw.println("  return "+c.getSimpleName()+"::getSerialVersionUID_s();"); 
      	pw.println("}"); 
      	pw.println(); 
      	pw.println("const char *"+c.getSimpleName()+"::getSerialID() const {");
      	pw.println("  return "+c.getSimpleName()+"::getSerialID_s();");
      	pw.println("}"); 
      	
      	if (hasSuper) {
      		
          	pw.println("const char *"+c.getSimpleName()+"::getSuperSerialID_s() {");
          	pw.println("  return "+this.getCxxClassName(c.getSuperclass())+"::getSerialID_s();"); 
          	pw.println("}"); 
          	pw.println("const char *"+c.getSimpleName()+"::getSuperSerialID() const {");
          	pw.println("  return "+this.getCxxClassName(c.getSuperclass())+"::getSerialID_s();");
          	pw.println("}"); 
      	}
      	
      	pw.println();
      	if (fieldCount > 0 || !hasSuper) {
      		pw.println("void "+c.getSimpleName()+"::writeFields(const Fancy::ObjPtr<Fancy::Serializer> &serializer) const {");
      		for (Map.Entry<String,FieldInfo> f : fieldNames.entrySet()) {
       		
      			String wfn = primitiveWriteFns.get(f.getValue());
      			if (wfn == null)
      				wfn = "writeObject";
       			
      			if (this.fancybaseAbi <= 0) {
      				pw.println("#if defined FANCYLIBS_ABI && FANCYLIBS_ABI >= 3");
      			}			
      			if (this.fancybaseAbi <= 0 || this.fancybaseAbi >= 3) {
      				
      				pw.print("  serializer->"+wfn+"(");
      				quoteCString(pw,f.getKey());
      				pw.println(",this->"+f.getKey()+");");
      			}
      			if (this.fancybaseAbi <= 0) {
      				pw.println("#else");
      			}
      			if (this.fancybaseAbi <= 2 ){
          			      				
      				pw.println("  serializer->"+wfn+"(this->"+f.getKey()+");");
      			}
     			if (this.fancybaseAbi <= 0) {
      				pw.println("#endif");
      			}
      		}
      		// write fields of base class.
      		if (hasSuper) {
      			pw.println("  " + this.getCxxClassName(c.getSuperclass()) + "::writeFields(serializer);");
      		}
      		pw.println("}"); 
      		pw.println(); 
      		pw.println("Fancy::Serializable::ptr_t "+c.getSimpleName()+"::readFields(const Fancy::ObjPtr<Fancy::Serializer> &serializer, int64_t serialVersionUID) {"); 

  			if (this.fancybaseAbi <= 0) {
  				pw.println("#if defined FANCYLIBS_ABI && FANCYLIBS_ABI >= 3");
  			}

  			// fancylibs3 ABI
  			if (this.fancybaseAbi <= 0 || this.fancybaseAbi >= 3) {

  				for (Map.Entry<String,FieldInfo> f : fieldNames.entrySet()) {
  					
  					List<Long> exclusions = fieldExclusionsByOSVUID.get(f.getKey());
  					
  					if (exclusions != null) {
  						
  						pw.print("  if (");
  						
  						for (int i=0; i<exclusions.size(); ++i) {
  							
  							Long id = exclusions.get(i);
  							if (i > 0) {
  								pw.print(" && ");
  							}
  							pw.print("serialVersionUID != INT64_C("+id+")");
  						}
  						
  						pw.println("  ) {");
  					}
  						
  					String rfn3;
  					Class<?> serialIDClass1 = null;
  					Class<?> serialIDClass2 = null;
  					
  					if (f.getValue().getType().isArray()) {
  						
  						rfn3 = primitiveArrayReadFns.get(f.getValue().getType().getComponentType());
  						
  						if (rfn3 == null) {
  							rfn3 = "readObjectArray";
  							serialIDClass1 = f.getValue().getType().getComponentType();
  						}
  					}
  					else {
          				
  						if (f.getValue().getType().equals(List.class)) {
  							rfn3 = "readObjectList";
  							serialIDClass1 = f.getValue().getGenericParameter(0);
  						}
          				else if (f.getValue().getType().equals(Map.class)) {
          					
          					rfn3 = "readObjectMap";
          					serialIDClass1 = f.getValue().getGenericParameter(0);      					
          					serialIDClass2 = f.getValue().getGenericParameter(1);      					
          				}
          				else {
          					rfn3 = primitiveReadFns.get(f.getValue().getType());
          				}
  						
  						if (rfn3 == null) {
  							rfn3 = "readObject";
  							
              				serialIDClass1 = f.getValue().getType();
  						}
          			}
         		
      				pw.print("  this->"+f.getKey()+" = serializer->"+rfn3+"(");
      				quoteCString(pw,f.getKey());
      				
      				if (serialIDClass1 != null) {
      					pw.print(",");
      					if (serialIDClass1.equals(Object.class)) {
      						pw.print("NULL");
      					}
      					else {
      						pw.print(this.getCxxClassName(serialIDClass1));
      						pw.print("::getSerialID_s()");
      					}
      				}
      				
      				if (serialIDClass2 != null) {
      					pw.print(",");
      					if (serialIDClass2.equals(Object.class)) {
      						pw.print("NULL");
      					}
      					else {
      						pw.print(this.getCxxClassName(serialIDClass2));
      						pw.print("::getSerialID_s()");
      					}
      				}

      				pw.println(");");
      				
      				if (exclusions != null) {
      					
      					pw.println("  }");
      				}
  				}
  			}
  			
  			if (this.fancybaseAbi <= 0) {
  				pw.println("#else");
  			}
  			// fancybase2 ABI
  			if (this.fancybaseAbi <= 2 ){
  				
  	  			for (Map.Entry<String,FieldInfo> f : fieldNames.entrySet()) {
  	         		
  	      			List<Long> exclusions = fieldExclusionsByOSVUID.get(f.getKey());
  	      			
  	      			if (exclusions != null) {
  	      				
  	      				pw.print("  if (");
  	      				
  	      				for (int i=0; i<exclusions.size(); ++i) {
  	      					
  	      					Long id = exclusions.get(i);
  	      					if (i > 0) {
  	      						pw.print(" && ");
  	          				}
  	      					pw.print("serialVersionUID != INT64_C("+id+")");
  	      				}
  	      				
  	      				pw.println("  ) {");
  	      			}
  	      				
  	      			String rfn2 = primitiveReadFns.get(f.getValue().getType());
  	      			if (rfn2 == null) {
  	      				rfn2 = "readObject";
  	      			}
  	      			
  	      			pw.println("  this->"+f.getKey()+" = serializer->"+rfn2+"();");
  	      			
  	      			if (exclusions != null) {
  	      				
  	      				pw.println("  }");
  	      			}
  	  			}
  			}
  			
  			if (this.fancybaseAbi <= 0) {
  				pw.println("#endif");
  			}

      		// read fields of base class.
      		if (hasSuper) {

      			pw.println("  int64_t superId;");

      			if (superSVUIDs != null) {
      			
      				boolean first = true;
      				
      				for (Map.Entry<Long,Long> superEntry : superSVUIDs.entrySet()) {
      					
      					if (first) {
          					pw.println("  if (serialVersionUID == INT64_C(" + superEntry.getKey() + ")) {" );
      					}
      					else {
          					pw.println("  else if (serialVersionUID == INT64_C(" + superEntry.getKey() + ")) {" );
      					}
      					pw.println("    superId = INT64_C("+superEntry.getValue()+");");
      					pw.println("  }");
      					
      					first = false;
      				}
      				
      				pw.println("  else {");
      			}
      			
      			pw.println("  superId = "+
      					this.getCxxClassName(c.getSuperclass())+"::getSerialVersionUID_s();");

      			if (superSVUIDs != null) {
      				pw.println("  }");
      			}
      			pw.println("  return " + this.getCxxClassName(c.getSuperclass()) + "::readFields(serializer,superId);");
      		}
      		else {
      			pw.println("  return Fancy::Serializable::ptr_t();");
      		}
      		pw.println("}"); 
      		pw.println(); 
      	}
      	
      	// go through fields and write getters/setters.
       	for (FieldInfo f : c.getDeclaredFields()) {
       		
       		if (Modifier.isStatic(f.getModifiers()) ||
       				Modifier.isTransient(f.getModifiers()))
       			continue;
       		
       		String getter = 
       			f.getType() == boolean.class ?
       			PropertyHelper.getBooleanGetterName(f.getName()):
                PropertyHelper.getGetterName(f.getName());
       					
       		String setter = PropertyHelper.getSetterName(f.getName());

  			pw.println(this.getRetVal(f.getType())+c.getSimpleName()+"::"+getter+"() const {");
       		pw.println("  return this->"+f.getName()+";");
       		pw.println("}");
       		pw.println();

    		pw.println("void "+c.getSimpleName()+"::"+setter+"("+this.getArgPfx(f.getType())+"v) {");
       		pw.println("  this->"+f.getName()+" = v;");
       		pw.println("}");
       		pw.println(); 		
       	}
       	    
       	// Add visitor accept methods.
       	for (MethodInfo m : c.getDeclaredMethods()) {
       		
       		if (!isVisitorAccept(m)) continue;
       		if (Modifier.isAbstract(m.getModifiers())) continue;
       		
       		pw.println("void "+c.getSimpleName()+"::accept("+this.getArgPfx(m.getParameterTypes()[0])+"v) {");
       		pw.println("  Fancy::ObjPtr<"+c.getSimpleName()+"> this_p(this);");
       		pw.println("  v->visit(this_p);");
       		pw.println("}");
       		pw.println();
       	}

       	// end of namespace(s).
    	for (int i=0;i<namespaceDepth;++i) {
    		pw.print("}");
    	}
    	pw.println();
    	pw.close();
    }
    
    private void gatherAllInterfaceMethods(ClassInfo c, List<MethodInfo> allInterfaceMethods, Set<Method> visitedMethods) {
    	
    	for (MethodInfo mi : c.getDeclaredMethods()) {
    		
    		if (visitedMethods.contains(mi.getMethod())) continue;
    		allInterfaceMethods.add(mi);
    		visitedMethods.add(mi.getMethod());
    	}
    	
    	Class<?>[] ifaces = c.getInterfaces();
    	
    	if (ifaces == null) return;
    	
    	for (Class<?> iface : ifaces) {
    		this.gatherAllInterfaceMethods(this.getClassInfo(iface),allInterfaceMethods,visitedMethods);
    	}
    }
    
    private void processInterface(ClassInfo c, boolean remote) throws FileNotFoundException, UnsupportedEncodingException, IllegalArgumentException, IllegalAccessException {
        
    	ISO8601Format fmt = new ISO8601Format(ISO8601Format.DATETIME_FORMAT);
    	boolean hasSuper = c.getInterfaces() != null && c.getInterfaces().length > 0;
    	
    	File cxxFile = this.getCxxFile(c.getClazz());
    	File headerFile = this.getHeaderFile(c.getClazz());
    	String defn = headerFile.getName().toUpperCase(Locale.ENGLISH).replace('.','_');
    	
    	this.getLog().info("Writing header of interface ["+c.getName()+"] to ["+headerFile+"].");
    	
    	PrintWriter pw = new PrintWriter(headerFile,"utf-8");
       	
    	pw.println("//"); 
    	pw.println("// File: " + headerFile.getName()); 
    	pw.println("// generated: " + fmt.format(Calendar.getInstance())); 
    	pw.println("// generator: "+this.generator+"."); 
       	pw.println("// Note: Use 'mvn process-sources' to regenerate this file."); 
       	pw.println("//"); 
       	pw.println();
       	pw.println("#ifndef "+defn);
       	pw.println("#define "+defn+" 1");
       	pw.println();
       	
       	if (this.abiInclude != null)
       		pw.println("#include <"+this.abiInclude+">"); 
       	
       	if (hasSuper) {
            
            for (Class<?> si : c.getInterfaces())
                pw.println("#include <"+this.getInclude(si)+">"); 
       	}
       	else {
       		pw.println("#include <fancyobject.h>"); 
       	}
       	
       	pw.println();
       	
       	SortedMap<String,SortedSet<String>> fwdDeclsByNamespace = new TreeMap<String, SortedSet<String>>();
       	
       	if (remote)
       		addFwdDecl(fwdDeclsByNamespace,"Fancy","class ReflectionTarget");
       	
       	// Add fwd decls for method parameters.
       	for (MethodInfo m : c.getDeclaredMethods()) {
       		
       		if (!Modifier.isPublic(m.getModifiers())) continue;
       		
       		for (Class<?> pc : m.getParameterTypes()) {
       			
       			this.addFwdDecl(fwdDeclsByNamespace,pc);
       		}

   			this.addFwdDecl(fwdDeclsByNamespace,m.getReturnType());
       	}
       	
       	for(Map.Entry<String,SortedSet<String>> fwdDecl : fwdDeclsByNamespace.entrySet()) {
       		
       		int n = this.writeNamespace(pw,fwdDecl.getKey());
       		for (String decl : fwdDecl.getValue()) {
       			pw.println(decl+";");
       		}
       		for (int i=0;i<n;++i) {
        		pw.print("}");
        	}
       		pw.println();
       	}
       	
       	pw.println();
       	
       	String clsDecl;
       	
       	if (this.abiTag != null)
       		clsDecl = this.abiTag + " " + c.getSimpleName();
       	else
       		clsDecl = c.getSimpleName();
       	
       	int namespaceDepth = this.writeNamespace(pw,c.getClazz());
       	
       	printJavaDoc(pw,0,c.getJavaDoc());
       		
       	if (hasSuper) {
       		pw.print("class "+clsDecl);

            for (int i=0;i<c.getInterfaces().length;++i) {

                pw.print(i>0?", ":": ");

                Class<?> si = c.getInterfaces()[i];
                pw.print("virtual public "+this.getCxxClassName(si));
            }

       		pw.println(" {");
       	} else {
       		pw.println("class "+clsDecl+" : virtual public Fancy::Object {");
       	    
       	}
       	
       	pw.println("protected:");
       	pw.println("  "+c.getSimpleName()+"();");
       	pw.println("  virtual ~"+c.getSimpleName()+"();");
       	pw.println(); 
       	pw.println("public:");
      	pw.println("  typedef Fancy::ObjPtr<"+c.getSimpleName()+"> ptr_t;");
      	pw.println();
      	
      	this.propagatePrimitiveFinalFields(pw,c);
      	
      	if (remote) {
      		pw.println("  static ptr_t newProxyInstance(const Fancy::ObjPtr<Fancy::ReflectionTarget> &target);");
      		pw.println("  static Fancy::Object::ptr_t newProxyObject(const Fancy::ObjPtr<Fancy::ReflectionTarget> &target);");
      		pw.println("  static Fancy::ObjPtr<Fancy::ReflectionTarget> reflect(const ptr_t &impl);");
      		pw.println("  static Fancy::ObjPtr<Fancy::ReflectionTarget> reflectObject(const Fancy::Object::ptr_t &obj);");
      		pw.println("  static const char *getSerialID_s();");
      	}
      	
       	// Add decls for methods.
       	for (MethodInfo m : c.getDeclaredMethods()) {
       		
       		if (!Modifier.isPublic(m.getModifiers())) continue;
       		
   			printJavaDoc(pw,2,m.getJavaDoc());

           	pw.print("  virtual "+this.getRetVal(m.getReturnType())+m.getName()+"(");
       		
       		Class<?>[] pt = m.getParameterTypes();
       		for (int i=0;i<pt.length;++i) {
       			if (i>0) pw.print(", ");
       			pw.print(this.getArgPfx(pt[i])+m.getParameterName(i));
       		}       		
       		pw.println(") = 0;");
       	}
       	
       	pw.println("};"); // end of class
    	
       	// end of namespace(s).
    	for (int i=0;i<namespaceDepth;++i) {
    		pw.print("}");
    	}
    	pw.println();

    	// set the ObjPtr template to the same ABI visibility than the class itself.
    	if (this.abiTag != null) {
    		pw.println("namespace Fancy {");
    		pw.println("  template class "+this.abiTag+" ObjPtr<"+this.getCxxClassName(c.getClazz())+">;");
    		pw.println("}");
    	}

    	pw.println("#endif // "+defn);
    	pw.close();
    	
    	// write implementation file.
    	this.getLog().info("Writing implementation of interface ["+c.getName()+"] to ["+cxxFile+"].");
    	
    	pw = new PrintWriter(cxxFile,"utf-8");
       	
    	pw.println("//"); 
    	pw.println("// File: " + cxxFile.getName()); 
    	pw.println("// generated: " + fmt.format(Calendar.getInstance())); 
    	pw.println("// generator: "+this.generator+"."); 
       	pw.println("// Note: Use 'mvn process-sources' to regenerate this file."); 
       	pw.println("//"); 
       	pw.println();
       	
       	SortedSet<String> includes = new TreeSet<String>();
       	
       	this.addInclude(includes,c.getClazz()); 
       	
   		List<MethodInfo> allInterfaceMethods = new ArrayList<MethodInfo>();

   		if (remote) {
       		includes.add("fancyreflectiontarget.h");
       		includes.add("fancynumbers.h");
       		includes.add("fancylist.h");
       		includes.add("fancyexception.h");
       		includes.add("fancybaselog.h");
       		includes.add("string.h");
       		
       		Set<Method> visitedMethods = new HashSet<Method>();
       		
       		this.gatherAllInterfaceMethods(c,allInterfaceMethods,visitedMethods);
       		
       		// Add decls for methods.
			for (MethodInfo m : allInterfaceMethods) {
				
				if (!Modifier.isPublic(m.getModifiers())) continue;
				
				Class<?>[] pt = m.getParameterTypes();
				for (int i = 0; i < pt.length; ++i) {
					this.addInclude(includes,pt[i]);
				}
				this.addInclude(includes,m.getReturnType());
			}
   		}
       	
       	for(String include : includes) {
       		pw.println("#include <"+include+">");
       	}
   		pw.println();
       	
       	namespaceDepth = this.writeNamespace(pw,c.getClazz());
       	
       	// ctor
       	pw.println(c.getSimpleName()+"::"+c.getSimpleName()+"()");
       	pw.println("{}");
       	pw.println();
       	// dtor
       	pw.println(c.getSimpleName()+"::~"+c.getSimpleName()+"()");
      	pw.println("{}");
       	pw.println(); 
       	
		if (remote) {
			// declare proxy class.
			String proxyCls = c.getSimpleName() + "__reflection_proxy";
			writeHiddenClassHeader(pw);
			pw.println(proxyCls + " : public " + c.getSimpleName() + " {");
			pw.println("  Fancy::ObjPtr<Fancy::ReflectionTarget> target;");
			pw.println("public:");
			pw
					.println("  "
							+ proxyCls
							+ "(const Fancy::ObjPtr<Fancy::ReflectionTarget> &_target);");
			pw.println("  virtual ~" + proxyCls + "();");
			// Add decls for methods.
			for (MethodInfo m : allInterfaceMethods) {

				if (!Modifier.isPublic(m.getModifiers())) continue;
				
				pw.print("  " + this.getRetVal(m.getReturnType()) + m.getName()
						+ "(");

				Class<?>[] pt = m.getParameterTypes();
				for (int i = 0; i < pt.length; ++i) {
					if (i > 0)
						pw.print(", ");
					pw.print(this.getArgPfx(pt[i]) + m.getParameterName(i));
				}
				pw.println(");");
			}
			pw.println("};"); // end of proxy class.

			// implement proxy class.
			pw
					.println(proxyCls
							+ "::"
							+ proxyCls
							+ "(const Fancy::ObjPtr<Fancy::ReflectionTarget> &_target)");
			pw.println(" : target(_target) {}");
			pw.println(proxyCls + "::~" + proxyCls + "() {}");

			for (MethodInfo m : allInterfaceMethods) {

				if (!Modifier.isPublic(m.getModifiers())) continue;
				
				pw.print(this.getRetVal(m.getReturnType()) + proxyCls + "::"
						+ m.getName() + "(");

				Class<?>[] pt = m.getParameterTypes();
				for (int i = 0; i < pt.length; ++i) {
					if (i > 0)
						pw.print(", ");
					pw.print(this.getArgPfx(pt[i]) + m.getParameterName(i));
				}
				pw.println(") {");

				if (pt.length > 0)
					pw.println("  Fancy::ObjectArray<Fancy::Object>::ptr_t args = Fancy::ObjectArray<Fancy::Object>::newInstance("
							+ pt.length + ");");
				else
					pw.println("  Fancy::ObjectArray<Fancy::Object>::ptr_t args;");
				
				for (int i = 0; i < pt.length; ++i) {

					String pToObj = primitiveToObjs.get(pt[i]);
					if (pToObj == null)
						pw.println("  args->elem("+i+") = " + m.getParameterName(i) + ";");
					else
						pw.println("  args->elem("+i+") = " + pToObj + "(" + m.getParameterName(i)
								+ ");");
				}

				String pFromObj = primitiveFromObjs.get(m.getReturnType());

				if (pFromObj != null)
					pw.println("  return " + pFromObj
							+ "(this->target->invokeMethod(\"" + m.getName()
							+ "\",args));");
				else if (m.getReturnType() == void.class)
					pw.println("  this->target->invokeMethod(\"" + m.getName()
							+ "\",args);");
				else
					pw.println("  return this->target->invokeMethod(\"" + m.getName()
							+ "\",args);");

				pw.println("}");
			}
			// end of proxy class implementation.

			pw.println(c.getSimpleName()+ "::ptr_t "
							+ c.getSimpleName()
							+ "::newProxyInstance(const Fancy::ObjPtr<Fancy::ReflectionTarget> &target) {");
			pw.println("  return new " + proxyCls + "(target);");
			pw.println("}");
			
			pw.println("Fancy::Object::ptr_t "
					+ c.getSimpleName()
					+ "::newProxyObject(const Fancy::ObjPtr<Fancy::ReflectionTarget> &target) {");
			pw.println("  return new " + proxyCls + "(target);");
			pw.println("}");
	
			// declare sorted methods array
			SortedMap<String, MethodInfo> sortedMethods = new TreeMap<String, MethodInfo>();
			for (MethodInfo m : allInterfaceMethods) {
				
				if (!Modifier.isPublic(m.getModifiers())) continue;
				
				sortedMethods.put(m.getName(), m);
			}

			pw.println("static const char * const "+c.getSimpleName()+"__methods[] = {");
			for (String mn : sortedMethods.keySet()) {
				pw.println("  \"" + mn + "\",");
			}
			pw.println("  NULL");
			pw.println("};");
			
			for (String mn : sortedMethods.keySet()) {
				MethodInfo m = sortedMethods.get(mn);
				
				pw.println("static const char * const "+c.getSimpleName()+"__"+mn+"__param_types[] = {");
				
				Class<?>[] pt = m.getParameterTypes();
				for (int i=0;i<pt.length;++i) {
					pw.println("  \""+pt[i].getName()+"\",");
				}
				pw.println("  NULL");
				pw.println("};");
			}

			// declare reflection target class.
			String reflectCls = c.getSimpleName() + "__reflection_target";
			writeHiddenClassHeader(pw);
			pw.println(reflectCls + " : public Fancy::ReflectionTarget {");
			pw.println("  " + c.getSimpleName() + "::ptr_t impl;");

			pw.println("  static int getMethodID(const char *mn) {");
			pw.println("    int i0=0;");
			pw.println("    int i1=" + sortedMethods.size() + ";");
			pw.println("    while (i0<i1) {");
			pw.println("      int i = i0 + (i1-i0)/2;");
			pw.println("      int r =::strcmp(mn,"+c.getSimpleName()+"__methods[i]);");
			pw.println("      if (r == 0) return i;");
			pw.println("      if (r < 0) i1=i;");
			pw.println("      else       i0=i+1;");
			pw.println("    }");
			pw.println("    if (i0 < " + sortedMethods.size()
					+ " && 0==::strcmp(mn,"+c.getSimpleName()+"__methods[i0])) return i0;");
			pw.println("    return -1;");
			pw.println("  }");
			pw.println("public:");
			pw.println("  " + reflectCls + "(const " + c.getSimpleName()
					+ "::ptr_t &_impl);");
			pw.println("  virtual ~" + reflectCls + "();");
			pw.println("  Fancy::Object::ptr_t invokeMethod(const char *method, const Fancy::ObjectArray<Fancy::Object>::ptr_t &args);");
			pw.println("  const char * const *enumMethods(int *nMethodsReturn) const;");
			pw.println("  const char * const *enumMethodParameterTypes(const char *method, int *nParamsReturn) const;");
			pw.println("  const char *getMethodReturnType(const char *method) const;");
			pw.println("  const char *getSerialID() const;");
			pw.println("};"); // end of reflection target class.

			// implement reflection target class.
			pw.println(reflectCls + "::" + reflectCls + "(const "
					+ c.getSimpleName() + "::ptr_t &_impl)");
			pw.println(" : impl(_impl) {}");
			pw.println(reflectCls + "::~" + reflectCls + "() {}");

			pw.println("const char *"+reflectCls+"::getSerialID() const {");
			pw.println("  return "+c.getSimpleName()+"::getSerialID_s();");
			pw.println("}");

			pw.println("Fancy::Object::ptr_t "
							+ reflectCls
							+ "::invokeMethod(const char *method, const Fancy::ObjectArray<Fancy::Object>::ptr_t &args) {");
			pw.println("  FANCY_OBJ_PTR_CHECK(this->impl);");
			pw.println("  switch(getMethodID(method)) {");

			int im = 0;
			for (String mn : sortedMethods.keySet()) {

				MethodInfo m = sortedMethods.get(mn);
				Class<?>[] pt = m.getParameterTypes();

				pw.println("  case " + im + ":");
				if (pt.length != 0) {
					pw.println("    if (args.isNull() || args->getSize() != " + pt.length + ")");
					pw.println("      throw Fancy::Exception(FANCYBASE_LOG_APP,LOG_ERR,\"Invalid number of arguments specified in call to "
							+ c.getSimpleName() + "::" + mn + "\");");
				}
				String pToObj = primitiveToObjs.get(m.getReturnType());
				if (pToObj != null)
					pw.println("    return " + pToObj + "(this->impl->" + mn
							+ "(");
				else if (m.getReturnType() == void.class)
					pw.println("    this->impl->" + mn + "(");
				else
					pw.println("    return this->impl->" + mn + "(");

				for (int i = 0; i < pt.length; ++i) {

					String delim = i < pt.length - 1 ? "," : "";

					String pFromObj = primitiveFromObjs.get(pt[i]);
					if (pFromObj == null)
						pw.println("      args->elem("+i+")" + delim);
					else
						pw.println("      " + pFromObj + "(args->elem("+i+"))" + delim);
				}
				if (pToObj == null)
					pw.println("    );");
				else
					pw.println("    ));");

				if (m.getReturnType() == void.class)
					pw.println("    return Fancy::Object::ptr_t();");
				++im;
			}
			pw.println("  default:");
			pw.println("    throw Fancy::Exception(FANCYBASE_LOG_APP,LOG_ERR,\"Invalid method %s specified in call to interface "
							+ c.getSimpleName() + ".\",method);");

			pw.println("  }");
			pw.println("}");

			pw.println("const char * const *"
							+ reflectCls
							+ "::enumMethods(int *nMethodsReturn) const {");
			pw.println("  if (nMethodsReturn) *nMethodsReturn="+sortedMethods.size()+";");
			pw.println("  return "+c.getSimpleName()+"__methods;");
			pw.println("}");

			pw.println("const char * const *"
							+ reflectCls
							+ "::enumMethodParameterTypes(const char *method, int *nParamsReturn) const {");
			
			pw.println("  switch(getMethodID(method)) {");

			im = 0;
			for (String mn : sortedMethods.keySet()) {

				MethodInfo m = sortedMethods.get(mn);
				Class<?>[] pt = m.getParameterTypes();

				pw.println("  case " + im + ":");
				pw.println("    if (nParamsReturn) *nParamsReturn="+pt.length+";");
				pw.println("    return "+c.getSimpleName()+"__"+mn+"__param_types;");
				
				++im;
			}
			pw.println("  default:");
			pw.println("    throw Fancy::Exception(FANCYBASE_LOG_APP,LOG_ERR,\"Invalid method %s specified in query of interface "
							+ c.getSimpleName() + ".\",method);");

			pw.println("  }");
			pw.println("}");

			pw.println("const char *"+reflectCls+"::getMethodReturnType(const char *method) const {");

			pw.println("  switch(getMethodID(method)) {");

			im = 0;
			for (String mn : sortedMethods.keySet()) {

				MethodInfo m = sortedMethods.get(mn);
				
				pw.println("  case " + im + ":");
				pw.println("    return \""+m.getReturnType().getName()+"\";");
				
				++im;
			}

			pw.println("  default:");
			pw.println("    throw Fancy::Exception(FANCYBASE_LOG_APP,LOG_ERR,\"Invalid method %s specified in query of interface "
							+ c.getSimpleName() + ".\",method);");

			pw.println("  }");
			pw.println("}");
			
			// end of reflection target class implementation.

			pw.println("Fancy::ObjPtr<Fancy::ReflectionTarget> "
					+ c.getSimpleName() + "::reflect(const "
					+ c.getSimpleName() + "::ptr_t &impl) {");
			pw.println("  return new " + reflectCls + "(impl);");
			pw.println("}");
			pw.println("Fancy::ObjPtr<Fancy::ReflectionTarget> "
					+ c.getSimpleName() + "::reflectObject(const Fancy::Object::ptr_t &obj) {");
			pw.println("  "+ c.getSimpleName() + "::ptr_t impl = obj.check_cast<"+ c.getSimpleName() +">();");
			pw.println("  return new " + reflectCls + "(impl);");
			pw.println("}");
      		pw.println("const char *" + c.getSimpleName() + "::getSerialID_s() {");
      		pw.println("  return \"" + c.getName() + "\";");
      		pw.println("}");
		}
       	
       	// end of namespace(s).
    	for (int i=0;i<namespaceDepth;++i) {
    		pw.print("}");
    	}
    	pw.println();
    	pw.close();
    }
	
    private void processEnumClass(ClassInfo c) throws FileNotFoundException, UnsupportedEncodingException, IllegalArgumentException, IllegalAccessException {
        
    	ISO8601Format fmt = new ISO8601Format(ISO8601Format.DATETIME_FORMAT);
    	
     	File headerFile = this.getHeaderFile(c.getClazz());
     	File cxxFile = this.getCxxFile(c.getClazz());
     	String defn = headerFile.getName().toUpperCase(Locale.ENGLISH).replace('.','_');
    	
    	this.getLog().info("Writing header of enum ["+c.getName()+"] to ["+headerFile+"].");
    	
    	PrintWriter pw = new PrintWriter(headerFile,"utf-8");
       	
    	pw.println("//"); 
    	pw.println("// File: " + headerFile.getName()); 
    	pw.println("// generated: " + fmt.format(Calendar.getInstance())); 
    	pw.println("// generator: "+this.generator+"."); 
       	pw.println("// Note: Use 'mvn process-sources' to regenerate this file."); 
       	pw.println("//"); 
       	pw.println();
       	pw.println("#ifndef "+defn);
       	pw.println("#define "+defn+" 1");
       	pw.println();
       	if (this.abiInclude != null)
       		pw.println("#include <"+this.abiInclude+">"); 

       	pw.println("#include <fancyenum.h>");
       	pw.println("#include <string>");
       	pw.println();
       	
       	String clsDecl;
       	
       	int namespaceDepth = this.writeNamespace(pw,c.getClazz());
       	
       	EnumConstantInfo[] ec=c.getEnumConstants();
       	
      	if (this.abiTag != null)
       		clsDecl = this.abiTag + " " + c.getSimpleName();
       	else
       		clsDecl = c.getSimpleName();
       	
      	printJavaDoc(pw,0,c.getJavaDoc());

      	pw.println("class "+clsDecl+" : virtual public Fancy::Enum {");
       		
  
      	pw.println("public:");
      	pw.println("  typedef Fancy::ObjPtr<"+c.getSimpleName()+"> ptr_t;");
      	pw.println("  enum enum_t {");
      	{
      		int i;
      		for (i=0;i<ec.length-1;++i) {

          		if (ec[i].getJavaDoc() != null)
          			GenCxxMojo.printJavaDoc(pw,4,ec[i].getJavaDoc());
          		
      			pw.println("    "+ec[i].getConstant()+",");
      		}
      		
      		if (ec[i].getJavaDoc() != null)
      			GenCxxMojo.printJavaDoc(pw,4,ec[i].getJavaDoc());
      		
      		pw.println("    "+ec[i].getConstant());
       	}
       	pw.println("  };"); // end of enum_t
       	
       	pw.println("  static ptr_t newInstance(enum_t e);");
       	pw.println("  static ptr_t newInstance(const char *s);");
       	pw.println("  static ptr_t newInstance(const std::string &s) {");
       	pw.println("    return "+c.getSimpleName()+"::newInstance(s.c_str());");
       	pw.println("  }");
       	pw.println("  static Fancy::Serializable::ptr_t newSerializableInstance();");
       	pw.println("  static const char *getSerialID_s();");
       	pw.println("  const char *getSerialID() const;");
       	pw.println("  int32_t getMaxValue() const;");
       	pw.println("  std::string toString() const;");
       	pw.println("private:");
       	pw.println("  "+c.getSimpleName()+"(enum_t e);");
       	pw.println("  virtual ~"+c.getSimpleName()+"();");
       	pw.println("  Fancy::Serializable::ptr_t resolveRead(int32_t v) const;");

      	pw.println("};"); // end of class
    	
      	// end of namespace(s).
    	for (int i=0;i<namespaceDepth;++i) {
    		pw.print("}");
    	}
    	pw.println();

    	// set the ObjPtr template to the same ABI visibility than the class itself.
    	if (this.abiTag != null) {
    		pw.println("namespace Fancy {");
    		pw.println("  template class "+this.abiTag+" ObjPtr<"+this.getCxxClassName(c.getClazz())+">;");
    		pw.println("}");
    	}
    	
    	pw.println("#endif // "+defn);
    	pw.close();
    	
    	// write implementation file.
    	this.getLog().info("Writing implementation of enum ["+c.getName()+"] to ["+cxxFile+"].");
    	
    	pw = new PrintWriter(cxxFile,"utf-8");
       	
    	pw.println("//"); 
    	pw.println("// File: " + cxxFile.getName()); 
    	pw.println("// generated: " + fmt.format(Calendar.getInstance())); 
    	pw.println("// generator: "+this.generator+"."); 
       	pw.println("// Note: Use 'mvn process-sources' to regenerate this file."); 
       	pw.println("//"); 
       	pw.println();
       	
       	pw.println("#include <"+headerFile.getName()+">"); 
       	pw.println("#include <fancyexception.h>");
       	pw.println("#include <fancybaselog.h>");
       	pw.println("#include <string.h>");
       	pw.println();
       	namespaceDepth = this.writeNamespace(pw,c.getClazz());
       	pw.println();
       	pw.println(c.getSimpleName()+"::ptr_t "+c.getSimpleName()+"::newInstance(enum_t e) {");      	
       	pw.println("  static const "+c.getSimpleName()+"::ptr_t instances[] = {");
      	{
      		int i;
      		for (i=0;i<ec.length-1;++i)
      			pw.println("    "+c.getSimpleName()+"::ptr_t(new "+c.getSimpleName()+"("+ec[i].getConstant()+")),");
      		
      		pw.println("    "+c.getSimpleName()+"::ptr_t(new "+c.getSimpleName()+"("+ec[i].getConstant()+"))");
       	}
       	pw.println("  };");
       	pw.println("  return instances[e];");
       	pw.println("}");
       	pw.println();
       	pw.println("Fancy::Serializable::ptr_t "+c.getSimpleName()+"::newSerializableInstance() {");
       	pw.println("  return "+c.getSimpleName()+"::newInstance("+ec[0].getConstant()+");");
       	pw.println("}");
       	pw.println();
       	pw.println("const char *"+c.getSimpleName()+"::getSerialID_s() {");
       	pw.println("  return \""+c.getName()+"\";");
       	pw.println("}");
       	pw.println();
       	pw.println("const char *"+c.getSimpleName()+"::getSerialID() const {");
       	pw.println("  return "+c.getSimpleName()+"::getSerialID_s();");
       	pw.println("}");
       	pw.println();
       	pw.println("int32_t "+c.getSimpleName()+"::getMaxValue() const {");
       	pw.println("  return "+(ec.length-1)+";");
       	pw.println("}");
       	pw.println();
       	pw.println("static const char * const static_enum_names[] = {");
      	{
      		int i;
      		for (i=0;i<ec.length-1;++i)
      			pw.println("    \""+ec[i].getConstant()+"\",");
      		
      		pw.println("    \""+ec[i].getConstant()+"\"");
       	}
       	pw.println("};");
       	pw.println("std::string "+c.getSimpleName()+"::toString() const {");
       	pw.println("  return static_enum_names[this->getValue()];");       	
       	pw.println("}");
       	pw.println();
       	pw.println(c.getSimpleName()+"::ptr_t "+c.getSimpleName()+"::newInstance(const char *s) {");      	
       	pw.println("  for (int32_t i=0;i<"+ec.length+";++i) {");
       	pw.println("    if (strcmp(static_enum_names[i],s)==0)");
       	pw.println("      return "+c.getSimpleName()+"::newInstance((enum_t)i);");
       	pw.println("  }");
		pw.println("  throw Fancy::Exception(FANCYBASE_LOG_APP,LOG_ERR,\"Invalid enum string [%s] for type ["+
				c.getSimpleName() +"] specified.\",s);");
       	pw.println("}");
       	pw.println();
       	pw.println(c.getSimpleName()+"::"+c.getSimpleName()+"(enum_t e) : Fancy::Enum((int32_t)e) {}");
       	pw.println();
       	pw.println(c.getSimpleName()+"::~"+c.getSimpleName()+"() {}");
       	pw.println();
       	pw.println("Fancy::Serializable::ptr_t "+c.getSimpleName()+"::resolveRead(int32_t v) const {");
       	pw.println("  return "+c.getSimpleName()+"::newInstance((enum_t)v);");       	
       	pw.println("}");
       	pw.println();

       	// end of namespace(s).
    	for (int i=0;i<namespaceDepth;++i) {
    		pw.print("}");
    	}
    	pw.println();
      	pw.close();
    }

    /* (non-Javadoc)
	 * @see org.apache.maven.plugin.Mojo#execute()
	 */
	@SuppressWarnings({"rawtypes" })
	public void execute() throws MojoExecutionException, MojoFailureException {
			
		if (this.generator == null) {
			try {
				Properties pp = new Properties();
				pp.load(this.getClass().getClassLoader().getResourceAsStream("META-INF/maven/org.clazzes.remoting/maven-remoting-plugin/pom.properties"));
			
				this.generator = pp.get("groupId") + "." + pp.get("artifactId") + "-" + pp.get("version");
			
			} catch (IOException e) {
				throw new MojoExecutionException("Plugin misconfiguration, resource [META-INF/maven/org.clazzes/maven-remoting-plugin/pom.properties] could not be found.",e);
			}
		}
		
		if (!this.targetDirectory.exists())
			throw new MojoFailureException("Target directory ["+this.targetDirectory+"] does not exist.");
			
		Set<Class> cp = new HashSet<Class>();
		Set<Class> remoteIfs=new HashSet<Class>();
		
		if (this.classes != null) {
			for (String clazzName : this.classes) {
				try {
					
					Class cls = Class.forName(clazzName);
					cp.add(cls);
					if (Remote.class.isAssignableFrom(cls))
						remoteIfs.add(cls);
						
				} catch (ClassNotFoundException e) {
					
					throw new MojoExecutionException("Class ["+clazzName+"] could not be found.",e);
				}
			}
		}
		
		if (this.packageClasses != null) {
			for (String clazzName : this.packageClasses) {
				
				try {
					Class<?> pkgCls = Class.forName(clazzName);
					
					List<Class> pkgClses = PackageHelper.getPackageClasses(pkgCls);
					
					for (Class cls:pkgClses) {
						cp.add(cls);
						if (Remote.class.isAssignableFrom(cls))
							remoteIfs.add(cls);
					}
					
				} catch (ClassNotFoundException e) {
					
					throw new MojoExecutionException("Class ["+clazzName+"] could not be found.",e);
				}
			}	
		}				

		if (this.remoteInterfaces != null) {
			for (String clazzName : this.remoteInterfaces) {
				try {
					
					Class cls = Class.forName(clazzName);
					cp.add(cls);
					remoteIfs.add(cls);
						
				} catch (ClassNotFoundException e) {
					
					throw new MojoExecutionException("Remote interface ["+clazzName+"] could not be found.",e);
				}
			}
		}
		
		for (Class cls : cp) {
			
			try {
				if (cls.isInterface()) {
					this.processInterface(this.getClassInfo(cls),remoteIfs.contains(cls));
				}
				else if (cls.isEnum()) {
					this.processEnumClass(this.getClassInfo(cls));
				}
				else {
					
					if (cls.getSuperclass() == null) continue;
				
					// we skip classes with no fields and no base class other than java.lang.Object.
					// this covers e.g. Visitor support classes.
					if (cls.getSuperclass() == Object.class && cls.getDeclaredFields().length == 0)
						continue;
					
					this.processPOJOClass(this.getClassInfo(cls));
				}
			} catch (Exception e) {
				throw new MojoExecutionException("Error processing ["+cls+"].",e);
			}
		}
	}
}
